module BruteForceSearches

    use iso_c_binding
    use Globals

    private

    !>  \brief  What a minimisable function should look like
    abstract interface
        function target_function(values,params) bind(c)
            import :: c_double, c_ptr   ! because we're in an abstract interface, we normally don't have access to all the stuff from the module, so we need to import
            type(c_ptr), value  ::  values
            type(c_ptr), value  ::  params
            real(kind=c_double) ::  target_function
        end function target_function
    end interface

    !>  \brief  Brute-force (exhaustive search) minimiser
    type, public    ::  BruteForceSearch
        private
        integer                                                 ::  number_of_dimensions = 0
        logical                                                 ::  is_in_memory = .false.
        procedure(target_function), pointer, nopass             ::  target_function => null()   !<  a pointer to the function the minimiser aims to minimise.
                                                                                                !!  no pass because there an assumption by default that type-bound procedures have an object of the type to which they are bound as first argument.
                                                                                                !!  this default assumption is overriden by no pass.
        type(c_ptr),    allocatable                             ::  parameters(:)               !<  pointer to object (usually fmc) with parameters needed by the function.
                                                                                                !!  When multithreading, each thread has a pointer to an area in memory.
        real(kind=8),   allocatable                             ::  starting_value(:)           !<  for each dimension, the starting value
        real(kind=8),   allocatable                             ::  best_value(:)               !<  for each dimension, the starting value
        real(kind=8),   allocatable                             ::  half_range(:)               !<  for each dimension, the half-range to search over
        real(kind=8),   allocatable                             ::  step_size(:)                !<  for each dimension, the stepping size during the search
        logical,        allocatable                             ::  dimension_at_max(:)         !<  for each dimension, whether the search has already reached the maximum
        real                                                    ::  best_score
        integer                                                 ::  num_iterations              !<  the total number of iterations needed to complete the exhaustive search
        logical                                                 ::  print_progress_bar = .true. !<  show a progress bar when running the search
        contains
            procedure,  public      ::  Init
            procedure,  public      ::  Deallocate
            procedure,  public      ::  IncrementCurrentValues
            procedure,  public      ::  Run
            procedure,  public      ::  GetBestValues
            procedure,  public      ::  GetBestValue
            procedure,  public      ::  GetBestScore
    end type

    contains

    !>  \brief  Return the best values found by the search
    function GetBestValues(self)
        class(BruteForceSearch),    intent(in)  ::  self
        real(kind=8),   allocatable             ::  GetBestValues(:)
        GetBestValues = self%best_value
    end function GetBestValues
    function GetBestValue(self,value_index)
        class(BruteForceSearch),    intent(in)  ::  self
        integer,                    intent(in)  ::  value_index
        real(kind=8)                            ::  GetBestValue
        GetBestvalue = self%best_value(value_index)
    end function GetBestValue
    function GetBestScore(self)
        class(BruteForceSearch),    intent(in)  ::  self
        real                                    ::  GetBestScore
        GetBestScore = self%best_score
    end function GetBestScore

    !>  \brief  Run the brute force minimiser's exhaustive search
    subroutine Run(self,expected_starting_score)
        !DIR$ OPTIMIZE: 2
        use DatesAndTimes
        use ProgressBars
#ifdef _OPENMP
        use omp_lib
#endif
        ! arguments
        class(BruteForceSearch),    target,     intent(inout)   ::  self    !<  brute-force minimiser (must be initialised)
        real,                       optional,   intent(in)      ::  expected_starting_score !<  useful for debugging
        ! private variables
        integer                             ::  ierr
        real(kind=8),   target, allocatable ::  current_values(:), current_values_private(:)
        logical                             ::  search_completed
        integer                             ::  current_iteration, num_iterations_completed
        real                                ::  seconds_since_start, seconds_per_iteration, seconds_to_finish
        integer                             ::  clock_count_at_start, clock_count_current,clock_counts_per_second
        logical,            parameter       ::  output_at_every_iteration   =   .false.
        real                                ::  current_score
        type(c_ptr)                         ::  starting_values_cptr
        type(c_ptr)                         ::  current_values_private_cptr
        type(dateandtime)                   ::  mytime
        type(ProgressBar)                   ::  my_progress_bar
        logical,            parameter       ::  debug = .false.
        logical                             ::  spawn_threads
        integer                             ::  thread_num
        ! start work

        ! Check the minimiser is ready
        if (.not. self%is_in_memory) then
            call this_program%TerminateWithFatalError('BruteForceSearch::Run','Call Init first!')
        endif

        ! Allocate memory for the current values
        allocate(current_values(self%number_of_dimensions),stat=ierr)
        if (ierr .ne. 0) call this_program%TerminateWithFatalError('BruteForceSearch::Run','Failed memory allocation')


        ! Start work

        ! The time when we start
        call system_clock(clock_count_at_start,clock_counts_per_second)

        ! The incoming values & score
        starting_values_cptr = c_loc(self%starting_value)
        self%best_score = self%target_function(starting_values_cptr,self%parameters(1))
        self%best_value = self%starting_value
        if (debug) write(*,'(a,f6.3)') '**debug(BruteForceSearch::Run): score on input = ', self%best_score

        ! Sanity check: is the score at the starting parameter values as expected?
        if (present(expected_starting_score)) then
            if (abs(expected_starting_score+self%best_score) .gt. 0.01) then
                write(*,'(a,2(f0.3,1x))') '**error(BruteForceSearch::Run): starting score not as expected: ', &
                                            self%best_score, expected_starting_score
                call this_program%TerminateWithFatalError('BruteForceSearch::Run','failed sanity check')
            endif
        endif

        ! Will we be spawning threads?
        spawn_threads = .false.
#ifdef _OPENMP
        if (debug) print *, '**debug(BruteForceSearch::Run): omp_in_parallel, omp_get_max_threads: ', omp_in_parallel(), omp_get_max_threads()
        if (.not. omp_in_parallel()) then
            if (omp_get_max_threads() .gt. 1) spawn_threads = .true.
        endif
#endif


        ! The starting point for the brute-force search
        current_values = -huge(1.0d0)
        self%dimension_at_max = .false.

        if (.not. output_at_every_iteration .and. self%print_progress_bar) call my_progress_bar%Begin(self%num_iterations)

        ! How many iterations have we completed?
        num_iterations_completed = 0


        !$omp   parallel default(shared) &
        !$omp   private(current_iteration, current_values_private,current_score,current_values_private_cptr,thread_num)

#ifdef _OPENMP
        ! were we given the correct number of parameter pointers?
        if (spawn_threads) then
            if (omp_get_num_threads() .ne. size(self%parameters)) then
                write(*,'(2(a,i0),a)') '**error(BruteForceSearch::Run): want to spawn ', omp_get_num_threads(), &
                                    ' but have ', size(self%parameters), ' sets of parameters '
                call this_program%TerminateWithFatalError('BruteForceSearch::Run','Bad number of parameter pointers')
            endif
        endif
        thread_num = omp_get_thread_num()
#else
        thread_num = 0
#endif

        ! allocate & initialise
        current_values_private = current_values

        ! make sure the c pointer to the current values is setup
        current_values_private_cptr = c_loc(current_values_private)

        !$omp barrier

        ! start the brute-force search iterations
        !$omp   do
        do current_iteration=1,self%num_iterations

            !$omp critical (bf_run_crit_incr)
            ! what values for the parameters should we try now?
            call self%IncrementCurrentValues(current_values,search_completed)
            if (search_completed) then
                call this_program%TerminateWithFatalError('BruteForceSearch::Run','failed sanity check')
            endif
            current_values_private = current_values
            !$omp end critical (bf_run_crit_incr)


            ! Try the current parameters by calling the scoring function
            if (debug) then
                write(*,'(a)',advance='no') '**debug(BruteForceSearch::Run): calling scoring function with values: '
                print *, current_values
            endif
            current_score =  self%target_function(current_values_private_cptr,self%parameters(thread_num+1))


            !$omp critical (bf_run_crit_best)
            num_iterations_completed = num_iterations_completed + 1
            ! if the score is the best we've seen so far, remember the values and the score
            if (current_score .lt. self%best_score) then
                self%best_score = current_score
                self%best_value = current_values_private
            endif
            ! Progress bar
            if (.not. output_at_every_iteration .and. self%print_progress_bar) call my_progress_bar%Update(num_iterations_completed)

            ! Output for the user
            if (output_at_every_iteration) then
                ! add a printout with a time estimate
                call system_clock(clock_count_current)
                seconds_since_start = real(clock_count_current - clock_count_at_start) / real(clock_counts_per_second)
                seconds_per_iteration = seconds_since_start / current_iteration
                seconds_to_finish = (self%num_iterations - current_iteration + 1) * seconds_per_iteration
                write(*,'(3a,i6,a,i6,a,f6.3,2a)') '**info(BruteForceSearch::Run): ', mytime%dateandtimeasstring(), &
                                                    ': after iteration ', &
                                                    current_iteration, ' of ', self%num_iterations, '. current best score : ',&
                                                    self%best_score, '. estimated time to finish: ', &
                                                    mytime%durationasstring(seconds_to_finish)
            endif
            if (debug) then
                write(*,'(a)',advance='no') '**debug(BruteForceSearch::Run): values: '
                write(*,*) current_values
                write(*,'(a,f0.4)') '**debug(BruteForceSearch::Run): score: ', current_score
            endif
            !$omp end critical (bf_run_crit_best)

        enddo ! end of loop over exhaustive search iterations
        !$omp end do

        if (allocated(current_values_private)) deallocate(current_values_private)

        !$omp end parallel

        if (.not. output_at_every_iteration .and. self%print_progress_bar) call my_progress_bar%Finish()

        ! Search is done
        if (debug) then
            write(*,'(2(a,i6),a)') '**debug(BruteForceSearch::Run): brute-force search completed in ', num_iterations_completed, &
                            ' iterations (predicted number of iterations was ', self%num_iterations, ')'
            print *, '**debug(BruteForceSearch::Run): best value = ', self%best_value
            print *, '**debug(BruteForceSearch::Run): best score = ', self%best_score
        endif

    end subroutine Run

    !>  \brief  In the BF loop, this should be called before the scoring function. Before the BF loop, set all values in current_values to -huge
    subroutine IncrementCurrentValues(self,current_values,search_is_now_completed)
        ! arguments
        class(BruteForceSearch),    intent(inout)   ::  self
        real(kind=8),               intent(inout)   ::  current_values(*)
        logical,                    intent(out)     ::  search_is_now_completed
        ! private variables
        integer                             ::  i
        !

        ! do the increment
        search_is_now_completed= .true.
        do i=1,self%number_of_dimensions
            ! if we haven't reached the max for this dimension, increment it, and reset all previous dimensions to their starting point
            if (.not. self%dimension_at_max(i)) then
                ! if we got here, it means our search is not over yet
                search_is_now_completed = .false.
                ! increment the ith dimension
                ! it could be that the ith dimension has a very large negative value, perhaps because the search hasn't even started yet
                if (current_values(i) .lt. self%starting_value(i) - self%half_range(i)) then
                    current_values(i) = self%starting_value(i) - self%half_range(i)
                else
                    current_values(i) = current_values(i) + self%step_size(i)
                endif
                ! reset all previous dimensions to their starting values, and reset their dimension_at_max flags to .false.
                if (i .gt. 1) then
                    current_values(1:i-1) = self%starting_value(1:i-1) - self%half_range(1:i-1)
                    if (any( .not. self%dimension_at_max(1:i-1))) then
                        call this_program%TerminateWithFatalError('BruteForceSearch::IncrementCurrentValues','failed sanity check')
                    endif
                    self%dimension_at_max(1:i-1) = .false.
                endif
                ! if the ith dimension has reached or gone over its max, set it at the max, and set its logical flag dimension_at_max to .true.
                if (current_values(i) .ge. self%starting_value(i) + self%half_range(i)) then
                    current_values(i) = self%starting_value(i) + self%half_range(i)
                    self%dimension_at_max(i) = .true.
                endif
                exit
            endif
        enddo

        do i=1,self%number_of_dimensions
            if (current_values(i) .lt. self%starting_value(i) - self%half_range(i)) then
                current_values(i) = self%starting_value(i) - self%half_range(i)
            endif
        enddo

    end subroutine IncrementCurrentValues


    !>  \brief  initialise a brute-force minimiser
    subroutine Init(self,function_to_minimise,parameters,num_dim,starting_value,half_range,step_size,print_progress_bar)
        ! arguments
        class(BruteForceSearch),    intent(inout)   ::  self                    !<  the brute-force minimiser to initialise / allocate
        procedure(target_function)                  ::  function_to_minimise    !<  the function to minimise
        type(c_ptr),                intent(in)      ::  parameters(:)           !<  parameters (a c pointer to an object, an array, whatever... ;
                                                                                !!  probably makes sense to have it point to a derived-type object that
                                                                                !!  contains useful stuff for the function to be minimised).
                                                                                !!  One pointer per thread needs to be provided

        integer,                    intent(in)      ::  num_dim                 !<  number of dimensions
        real(kind=8),               intent(in)      ::  starting_value(:)       !<  starting point for the minimiser
        real(kind=8),               intent(in)      ::  half_range(:)           !<  half-range to search over
        real(kind=8),               intent(in)      ::  step_size(:)            !<  step size of the exhaustive search
        logical,        optional,   intent(in)      ::  print_progress_bar      !<  print a progress bar during the search
        ! private variables
        integer                 ::  ierr
        real(kind=8)            ::  current_values(num_dim)
        logical                 ::  search_done
        logical,    parameter   ::  debug = .false.
        ! start work

        ! if already allocated, deallocate
        if (self%is_in_memory) call self%Deallocate()

        ! allocate memory
        self%number_of_dimensions = num_dim
        allocate(   self%starting_value(self%number_of_dimensions),self%half_range(self%number_of_dimensions), &
                    self%step_size(self%number_of_dimensions), &
                    self%best_value(self%number_of_dimensions),self%dimension_at_max(self%number_of_dimensions),stat=ierr)
        if (ierr .ne. 0) then
            write(*,'(a,i3)') '**error(BruteForceSearch::Init): memory allocation failed with error ', ierr
            call this_program%TerminateWithFatalError('BruteForceSearch::Init','failed memory allocation')
        endif

        ! check the incoming starting values, half range and step size have expected dimensionality
        if (size(starting_value) .ne. num_dim .or. size(half_range) .ne. num_dim .or. size(step_size) .ne. num_dim) then
            write(*,'(a,i5,a,3i5)') '**error(BruteForceSearch::Init): incoming arrays do not have expected size of ', &
                                     num_dim, ': ', size(starting_value), size(half_range), size(step_size)
            call this_program%TerminateWithFatalError('BruteForceSearch::Init','fatal error')
        endif

        ! copy starting values, half range and step size over
        self%starting_value = starting_value
        self%half_range = half_range
        self%step_size = step_size

        !
        self%dimension_at_max = .false.

        ! work out how many iterations the exhaustive search will take
        ! we will start at lowest value for each dimension, then go in step sizes, until we get to upper limit.
        ! we will also do extra iterations for the starting value and the upper limit
        self%num_iterations = 0
        current_values = -huge(1.0d0)
        do
            call self%IncrementCurrentValues(current_values,search_done)
            if (search_done) exit
            self%num_iterations = self%num_iterations + 1
        enddo
        !do i=1,self%n
        !    self%num_iterations = self%num_iterations * ceiling(self%half_range(i)*2/self%step_size(i))
        !enddo


        if (debug) then
            write(*,'(a,i0)') '**info(BruteForceSearch::Init): predicted number of steps in the exhaustive search = ', &
                                self%num_iterations
            write(*,'(a,i0)') '**info(BruteForceSearch::Init): number of dimensions = ', self%number_of_dimensions
            print *, '**debug(BruteForceSearch::Init): half-ranges = ', self%half_range
            print *, '**debug(BruteForceSearch::Init): step sizes  = ', self%step_size
        endif

        ! set up pointer to function to minimise
        self%target_function => function_to_minimise

        !
        self%parameters = parameters

        ! progress bar?
        self%print_progress_bar = .true.
        if (present(print_progress_bar)) self%print_progress_bar = print_progress_bar

        ! set the memory flag
        self%is_in_memory = .true.
    end subroutine Init

    !>  \brief  Deallocate a brute-force minimiser
    subroutine Deallocate(self)
        ! argument
        class(BruteForceSearch),    intent(inout)   ::  self    !<  brute-force minimiser to deallocate
        ! private variables

        ! start work
        if (.not. self%is_in_memory) then
            call this_program%TerminateWithFatalError('BruteForceSearch::Deallocate', 'Not in memory, cannot deallocate')
        endif
        ! deallocate component arrays
        deallocate(self%starting_value,self%half_range,self%step_size, self%best_value,self%dimension_at_max)
        ! reset pointer
        nullify(self%target_function)
        ! reset values
        self%number_of_dimensions = 0
        self%num_iterations = 0
        self%is_in_memory = .false.
    end subroutine Deallocate


end module
