!
! Copyright 2014 Howard Hughes Medical Institute
! All rights reserved
! Use is subject to Janelia Farm Research Campus Software Copyright 1.1
! license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
!
!
module ImageCTFComparisons
    use Images
    use ContrastTransferFunctions
    type ImageCTFComparison
        type(Image)                     ::  img
        type(ContrastTransferFunction)  ::  ctf
        real                            ::  pixel_size  !<  Angstroms
        logical                         ::  find_phase_shift = .false.
    end type

    logical,    parameter               ::  fit_squared_ctf = .false. ! whether to fit squared ctf or absolute ctf

    contains

    !>  \brief  Scoring function to be minimised
    function ctffind_objective_function(array_of_values,scoring_parameters) bind(c)
        use iso_c_binding
        use Units
        use ContrastTransferFunctions
        ! arguments
        type(c_ptr), value ::  array_of_values
        type(c_ptr), value ::  scoring_parameters
        ! result
        real(kind=c_double) ::  ctffind_objective_function
        ! private variables
        real(kind=8),               pointer ::  values_to_try(:)    !<  1. Defocus1; 2. Defocus2; 3. Astigmatism angle; 4. Additional phase shift
        type(ImageCTFComparison),   pointer ::  comparison_object
        type(ContrastTransferFunction)      ::  my_ctf
        logical,                parameter   ::  debug = .false.
        ! start work
        call c_f_pointer(scoring_parameters,comparison_object)
        if (comparison_object%find_phase_shift) then
            call c_f_pointer(array_of_values,values_to_try,[4])
        else
            call c_f_pointer(array_of_values,values_to_try,[3])
        endif
        ! Set defocus values & astigmatism
        my_ctf = comparison_object%ctf
        call my_ctf%SetDefocus( convert(real(values_to_try(1)),angstroms,pixels,comparison_object%pixel_size), &
                                convert(real(values_to_try(2)),angstroms,pixels,comparison_object%pixel_size), &
                                convert(real(values_to_try(3)),degrees,radians) )
        if (comparison_object%find_phase_shift) then
            call my_ctf%SetAdditionalPhaseShift(real(values_to_try(4)))
        endif
        ctffind_objective_function = - comparison_object%img%GetCorrelationWithCTF(my_ctf,fit_squared_ctf)
        if (debug) then
            write(*,'(a,3(f0.2,1x),f0.5)') '**debug(ctffind_objective_function): values, score: ', &
                                            values_to_try,ctffind_objective_function
        endif
    end function ctffind_objective_function

end module ImageCTFComparisons

program ctffind

    use Globals
    use ImageFiles
    use Images
    use BruteForceSearches
    use UserInputs
    use UserSuppliedParameters
    use EmpiricalDistributions
    use Histograms
    use ContrastTransferFunctions
    use ImageCTFComparisons
    use BruteForceSearches
    use ConjugateGradients
    use Units
    use StringManipulations
    use UsefulFunctions
    use NumericTextFiles
    use DatesAndTimes
    use ProgressBars
    use omp_lib
    use, intrinsic :: iso_c_binding

    implicit none

    ! User input object
    type(UserInput)                 ::  my_user_input

    ! The user will supply values for these
    type(UserSuppliedFilename)      ::  input_filename
    type(UserSuppliedLogical)       ::  input_is_a_movie
    type(UserSuppliedInteger)       ::  number_of_frames_to_average
    type(UserSuppliedFilename)      ::  output_diagnostic_filename
    type(UserSuppliedReal)          ::  pixel_size
    type(UserSuppliedReal)          ::  acceleration_voltage
    type(UserSuppliedReal)          ::  spherical_aberration
    type(UserSuppliedReal)          ::  amplitude_contrast
    type(UserSuppliedInteger)       ::  box_size
    type(UserSuppliedReal)          ::  minimum_resolution
    type(UserSuppliedReal)          ::  maximum_resolution
    type(UserSuppliedReal)          ::  minimum_defocus
    type(UserSuppliedReal)          ::  maximum_defocus
    type(UserSuppliedReal)          ::  defocus_search_step
    type(UserSuppliedReal)          ::  astigmatism_tolerance
    type(UserSuppliedLogical)       ::  find_additional_phase_shift
    type(UserSuppliedReal)          ::  minimum_additional_phase_shift
    type(UserSuppliedReal)          ::  maximum_additional_phase_shift
    type(UserSuppliedReal)          ::  additional_phase_shift_search_step

    ! Variables
    type(ImageFile)                 ::  input_file
    type(Image)                     ::  current_input_image
    type(Image)                     ::  current_power_spectrum
    type(Image)                     ::  resampled_power_spectrum
    type(Image)                     ::  average_spectrum
    type(Image)                     ::  temp_image
    type(Image)                     ::  current_input_image_square
    integer                         ::  current_micrograph_number
    integer                         ::  number_of_micrographs
    integer                         ::  number_of_movie_frames
    integer                         ::  current_input_location
    integer                         ::  current_output_location
    integer                         ::  current_first_frame_within_average
    integer                         ::  number_of_tiles_used
    integer                         ::  convolution_box_size
    real                            ::  distance_between_zeroes
    type(ContrastTransferFunction)  ::  current_ctf
    type(ImageCTFComparison), target::  comparison_object
    type(BruteForceSearch)          ::  brute_force_search
    real(kind=8)                    ::  bf_midpoint(4), bf_halfrange(4), bf_stepsize(4)
    real(kind=8),   allocatable     ::  cg_starting_point(:)
    real(kind=8),   allocatable     ::  cg_accuracy(:)
    type(ConjugateGradient)         ::  conjugate_gradient
    real                            ::  threshold
    real                            ::  estimated_astigmatism_angle
    type(c_ptr), allocatable        ::  bf_param_array(:)
    character(len=filename_max_len) ::  output_text_fn
    character(len=512)              ::  comment_lines(4)
    type(NumericTextFile)           ::  output_text
    type(NumericTextFile)           ::  output_text_avrot
    type(DateAndTime)               ::  my_date_and_time
    logical,    parameter           ::  taper_micrograph = .true.
    logical,    parameter           ::  find_astigmatism_angle_using_mirror = .true. ! this saves time, compared to searching for the angle at the same time as the defocus values
    logical,    parameter           ::  make_micrograph_square = .true.
    logical,    parameter           ::  subtract_background = .true.
    logical,    parameter           ::  debug = .false.
    logical                         ::  compute_extra_stats ! this is pretty slow but gives the somewhat useful estimate of how far Thon rings were detected, as well as 1D versions of the spectra
    logical                         ::  boost_ring_contrast
    real(kind=8),   parameter       ::  frc_significance_threshold = 0.75d0
    real,       parameter           ::  lowest_allowed_minimum_resolution = 50.0
    integer                         ::  current_frame_within_average
    real                            ::  values_to_write_out(7)
    real(kind=8),   allocatable     ::  spatial_frequency(:)
    real(kind=8),   allocatable     ::  rotational_average(:), rotational_average_astig(:), rotational_average_astig_fit(:)
    real(kind=8),   allocatable     ::  frc_of_fit(:), frc_of_fit_sigma(:)
    integer                         ::  micrograph_square_dimension
    type(ProgressBar)               ::  my_progress_bar
    integer                         ::  last_bin_with_good_fit
    real                            ::  average, sigma
    character(len=:),   allocatable ::  command_line_options(:)
    integer                         ::  number_of_search_dimensions
    logical                         ::  resampling_is_necessary


    ! Variables specifically for the old-school input
    logical                         ::  old_school_input
    character(len=200)              ::  temp_string
    integer                         ::  arg_status
    real                            ::  xmag, dstep
    integer                         ::  current_arg

    ! Other command-line options
    logical                         ::  amplitude_spectrum_input
    logical                         ::  filtered_amplitude_spectrum_input


    !!
    !! Start
    !!


    ! Initialise the program
    call this_program%Init('CTFFind','4.0.16','2015',command_line_options)


    ! Parse command line options if any
    old_school_input = .false.
    amplitude_spectrum_input = .false.
    filtered_amplitude_spectrum_input = .false.
    compute_extra_stats = .true.
    boost_ring_contrast = .true.
    if (allocated(command_line_options)) then
        do current_arg=1,size(command_line_options)
            if (StringsAreEqual(command_line_options(current_arg),'--old-school-input')) then
                old_school_input = .true.
            endif
            if (StringsAreEqual(command_line_options(current_arg),'--amplitude-spectrum-input')) then
                amplitude_spectrum_input = .true.
            endif
            if (StringsAreEqual(command_line_options(current_arg),'--filtered-amplitude-spectrum-input')) then
                filtered_amplitude_spectrum_input = .true.
            endif
            if (StringsAreEqual(command_line_options(current_arg),'--fast')) then
                compute_extra_stats = .false.
                boost_ring_contrast = .false.
            endif
        enddo
    endif


    ! Interact with the user
    call my_user_input%Init(this_program%program_name)
    if (old_school_input) then
        !
        ! The old way of interacting with users. To be deprecated (when Relion lets us).
        !
        read(*,'(a)')                           temp_string
                                                input_filename%value = temp_string
        read(*,'(a)')                           temp_string
                                                output_diagnostic_filename%value = temp_string
        read(*,*)   spherical_aberration%value, acceleration_voltage%value, &
                                                amplitude_contrast%value, xmag, dstep
        read(*,*)      box_size%value, minimum_resolution%value, maximum_resolution%value, &
                                                minimum_defocus%value, maximum_defocus%value, defocus_search_step%value, &
                                                astigmatism_tolerance%value

        ! If we are getting dAst = 0.0, which is the default in Relion, the user probably
        ! expects the ctffind3 behaviour, which is no restraint on astigmatism
        if (astigmatism_tolerance%value .eq. 0.0) astigmatism_tolerance%value = -100.0

        pixel_size%value = dstep * 10000.0 / xmag

        input_filename%set_by_user = .true.
        output_diagnostic_filename%set_by_user = .true.
        spherical_aberration%set_by_user = .true.
        acceleration_voltage%set_by_user = .true.
        amplitude_contrast%set_by_user = .true.
        pixel_size%set_by_user = .true.
        box_size%set_by_user = .true.
        minimum_resolution%set_by_user = .true.
        maximum_resolution%set_by_user = .true.
        minimum_defocus%set_by_user = .true.
        maximum_defocus%set_by_user = .true.
        defocus_search_step%set_by_user = .true.
        astigmatism_tolerance%set_by_user = .true.
        number_of_frames_to_average%set_by_user = .false.
        number_of_frames_to_average%value = 1
        find_additional_phase_shift%set_by_user = .false.
        find_additional_phase_shift%value = .false.
        minimum_additional_phase_shift%set_by_user = .false.
        minimum_additional_phase_shift%value = 0.0
        maximum_additional_phase_shift%set_by_user = .false.
        maximum_additional_phase_shift%value = 0.0
        additional_phase_shift_search_step%set_by_user = .false.
        additional_phase_shift_search_step%value = 0.0

        ! Input file is normally initialised during the user interaction to work out whether it's a stack of movie frames
        input_is_a_movie%value  =   .false.
        call input_file%Init(input_filename%value)
        if (input_file%GetStackSize() .gt. 1) then
            call this_program%TerminateWithFatalError('ctffind','Input stacks are only supported when using the modern UI')
        endif

        ! Check the input values are reasonable
        if (minimum_resolution%value .gt. lowest_allowed_minimum_resolution) then
            minimum_resolution%value = lowest_allowed_minimum_resolution
            write(*,'(a,f0.1,a)') 'Overriding minimum resolution for fitting to ',lowest_allowed_minimum_resolution,' A'
        endif

        ! Output for old-school users
        write(*,'(/a)') ' CS[mm], HT[kV], AmpCnst, XMAG, DStep[um]'
        write(*,'(f5.1,f9.1,f8.2,f10.1,f9.3/)') spherical_aberration%value, acceleration_voltage%value, &
                                                amplitude_contrast%value, xmag, dstep

    else
        !
        ! The normal way of interacting with users
        !

        input_filename              =   my_user_input%GetFilenameFromUser(  'Input image file name','Filename of input image',&
                                                                            'input_filename','input.mrc',.true.)
        ! Open the input file
        call input_file%Init(input_filename%value)
        if (input_file%GetStackSize() .gt. 1) then
            input_is_a_movie        =   my_user_input%GetLogicalFromUser(   'Input is a movie (stack of frames)', &
                                                                            'Answer yes if the input file is a stack of '//&
                                                                            'frames from a dose-fractionated movie. If not, '//&
                                                                            'each image will be processed separately', &
                                                                            'input_is_a_movie','no')
        else
            input_is_a_movie%value  =   .false.
        endif

        if (input_is_a_movie%value) then
            number_of_frames_to_average = my_user_input%GetIntegerFromUser( 'Number of frames to average together', &
                                                                            'If the number of electrons per frame is too low, '//&
                                                                            'there may be strong artefacts in the '//&
                                                                            'estimated power spectrum. This can be '//&
                                                                            'alleviated by averaging frames '//&
                                                                            'with each other in real space '//&
                                                                            'before computing Fourier transforms.', &
                                                                            'num_frames_to_average','1')
        else
            number_of_frames_to_average%value = 1
        endif

        ! Continue with user interaction
        output_diagnostic_filename  =   my_user_input%GetFilenameFromUser(  'Output diagnostic filename', &
                                                                            'Will contain experimental power spectrum '//&
                                                                            'and best fit', &
                                                                            'output_diag_filename','diagnostic_output.mrc',.false.)

        pixel_size                  =   my_user_input%GetRealFromUser(      'Pixel size','in Angstroms','pixel_size', &
                                                                            '1.0',min_value=0.0)

        acceleration_voltage        =   my_user_input%GetRealFromUser(      'Acceleration voltage', &
                                                                            'in kV','acceleration_voltage','300.0',min_value=0.0)
        spherical_aberration        =   my_user_input%GetRealFromUser(      'Spherical aberration', &
                                                                            'in mm','spherical_aberration','2.7',min_value=0.0)
        amplitude_contrast          =   my_user_input%GetRealFromUser(      'Amplitude contrast', &
                                                                            'fraction of amplitude contrast','amplitude_contrast', &
                                                                            '0.07',min_value=0.0,max_value=1.0)
        box_size                    =   my_user_input%GetIntegerFromUser(   'Size of power spectrum to compute', &
                                                                            'Size of the power spectrum (pixels) to compute', &
                                                                            'box_size','512',min_value=128)
        minimum_resolution          =   my_user_input%GetRealFromUser(      'Minimum resolution', &
                                                                            'Lowest resolution used for fitting CTF (Angstroms)', &
                                                                            'minimum_resolution','30.0', &
                                                                            min_value=0.0, &
                                                                            max_value=lowest_allowed_minimum_resolution)
        maximum_resolution          =   my_user_input%GetRealFromUser(      'Maximum resolution', &
                                                                            'Highest resolution used for fitting CTF (Angstroms)', &
                                                                            'maximum_resolution','5.0',min_value=0.0)
        minimum_defocus             =   my_user_input%GetRealFromUser(      'Minimum defocus', &
                                                                            'Positive values for underfocus. Lowest value to '//&
                                                                            'search over. (Angstroms)','minimum_defocus', &
                                                                            '5000.0')
        maximum_defocus             =   my_user_input%GetRealFromUser(      'Maximum defocus', &
                                                                            'Positive values for underfocus. Highest value to '//&
                                                                            'search over. (Angstroms)','maximum_defocus', &
                                                                            '50000.0',min_value=minimum_defocus%value)
        defocus_search_step         =   my_user_input%GetRealFromUser(      'Defocus search step', &
                                                                            'Step size for defocus search (Angstroms)', &
                                                                            'defocus_search_step','500.0',min_value=1.0)
        astigmatism_tolerance       =   my_user_input%GetRealFromUser(      'Expected (tolerated) astigmatism', &
                                                                            'Astigmatism values much larger than this will be '//&
                                                                            'penalised (Angstroms; set negative to remove this '//&
                                                                            'restraint)', &
                                                                            'astigmatism_tolerance','100.0')
        find_additional_phase_shift =   my_user_input%GetLogicalFromUser(   'Find additional phase shift?', &
                                                                            'Input micrograph was recorded using a phase plate '//&
                                                                            'with variable phase shift, which you want to find',&
                                                                            'find_additional_phase_shift','no')
        if (find_additional_phase_shift%value) then
            minimum_additional_phase_shift=my_user_input%GetRealFromUser(   'Minimum phase shift', &
                                                                            'Lower bound of the search for additional '//&
                                                                            'phase shift. Phase shift is of scattered electrons '//&
                                                                            'relative to unscattered electrons. In radians.',&
                                                                            'minimum_phase_shift','0.0',&
                                                                            min_value=-3.15,max_value=3.15)
            maximum_additional_phase_shift=my_user_input%GetRealFromUser(   'Maximum phase shift', &
                                                                            'Upper bound of the search for additional '//&
                                                                            'phase shift. Phase shift is of scattered electrons '//&
                                                                            'relative to unscattered electrons. In radians.',&
                                                                            'maximum_phase_shift','0.0',&
                                                                            min_value=minimum_additional_phase_shift%value+0.1,&
                                                                            max_value=3.15)

            additional_phase_shift_search_step=my_user_input%GetRealFromUser('Phase shift search step', &
                                                                            'Step size for phase shift search (radians)', &
                                                                            'phase_shift_search_step','0.2', &
                                                                            min_value=0.01 &
                                                                            )
        else
            minimum_additional_phase_shift%value = 0.0
            maximum_additional_phase_shift%value = 0.0
            additional_phase_shift_search_step%value = 0.0
        endif
    endif ! end of check for old-school input

    ! Done interacting with the user
    call my_user_input%UpdateDefaults()

    ! We want to time the runtime, not the user interaction
    call this_program%SetStartDate()

    ! Do some argument checking
    if (minimum_resolution%value .lt. maximum_resolution%value) then
        call this_program%TerminateWithFatalError('ctffind','Minimum resolution must be greater than maximum resolution')
    endif
    if (minimum_defocus%value .gt. maximum_defocus%value) then
        call this_program%TerminateWithFatalError('ctffind','Minimum defocus must less than maximum defocus')
    endif

    ! How many micrographs are we dealing with?
    if (input_is_a_movie%value) then
        ! we only support 1 movie per file
        number_of_movie_frames = input_file%GetStackSize()
        number_of_micrographs  = 1
    else
        number_of_movie_frames = 1
        number_of_micrographs  = input_file%GetStackSize()
    endif


    ! Print out information about input file
    call input_file%PrintInfo()

    ! Prepare the average spectrum image
    call average_spectrum%Allocate(dims=[box_size%value,box_size%value,1],in_real_space=.true.)

    ! Prepare the output text file
    output_text_fn = output_diagnostic_filename%value
    call FilenameReplaceExtension(output_text_fn,'txt')
    call output_text%Init(output_text_fn,OPEN_TO_WRITE,7)
    write(comment_lines(1),'(4a)') 'Output from CTFFind version ', this_program%program_version, &
                                        ', run on ', my_date_and_time%DateAndTimeAsString()
    write(comment_lines(2),'(3a,i0)') 'Input file: ', input_filename%value, '; Number of micrographs: ', number_of_micrographs
    write(comment_lines(3),'(a,f0.3,a,f0.1,a,f0.1,a,f0.2)') 'Pixel size: ', pixel_size%value, &
                '; acceleration voltage: ', acceleration_voltage%value, &
                '; spherical aberration: ', spherical_aberration%value, &
                '; amplitude contrast: ', amplitude_contrast%value
    write(comment_lines(4),'(a,i0,4(a,f0.1))') 'Box size: ', box_size%value, &
                '; min res: ', minimum_resolution%value, '; max. res.: ', maximum_resolution%value, &
                '; min. def.: ', minimum_defocus%value, '; max. def.: ', maximum_defocus%value

    call output_text%WriteCommentLine(comment_lines(1))
    call output_text%WriteCommentLine(comment_lines(2))
    call output_text%WriteCommentLine(comment_lines(3))
    call output_text%WriteCommentLine(comment_lines(4))
    call output_text%WriteCommentLine('Columns: #1 - micrograph number; #2 - defocus 1 [Angstroms]; #3 - defocus 2; '//&
                                      '#4 - azimuth of astigmatism; #5 - additional phase shift [radians]; '//&
                                      '#6 - cross correlation; #7 - spacing '//&
                                      '(in Angstroms) up to which CTF rings were fit successfully')

    ! Prepare a text file with 1D rotational average spectra
    call FilenameAddSuffix(output_text_fn,'_avrot')


    if (.not. old_school_input .and. number_of_micrographs .gt. 1) then
        write(*,'(/a,i0,a)') 'Will estimate the CTF parameters for ', number_of_micrographs, ' micrographs.'
        write(*,'(2a/)')     'Results will be written to this file: ', trim(adjustl(output_text%filename))
        call my_progress_bar%Begin(number_of_micrographs)
    endif


    ! Loop over micrographs
    do current_micrograph_number=1,number_of_micrographs
        if (old_school_input .or. number_of_micrographs .eq. 1) then
            write(*,'(/a,i0,a,i0)') 'Working on micrograph ', current_micrograph_number, ' of ', number_of_micrographs
        endif

        number_of_tiles_used = 0
        average_spectrum = 0.0e0
        average_spectrum%is_in_real_space = .true.

        if (amplitude_spectrum_input .or. filtered_amplitude_spectrum_input) then
            call current_power_spectrum%ReadFromDisk(input_file,current_micrograph_number)
            call current_power_spectrum%ForwardFFT()
            call average_spectrum%Allocate(dims=[box_size%value,box_size%value,1],in_real_space=.false.)
            call current_power_spectrum%ClipInto(average_spectrum)
            call average_spectrum%BackwardFFT()
        else
            do current_first_frame_within_average=1,number_of_movie_frames,number_of_frames_to_average%value
                do current_frame_within_average=1,number_of_frames_to_average%value
                    current_input_location = current_first_frame_within_average &
                                           + number_of_movie_frames * (current_micrograph_number-1) &
                                           + (current_frame_within_average-1)
                    if (current_input_location .gt. number_of_movie_frames*current_micrograph_number) cycle
                    call current_input_image%ReadFromDisk(input_file,current_input_location)
                    if (current_input_image%IsConstant()) then
                        write(*,'(a,i0,3a)') '**error(ctffind): location ', current_input_location, ' of input file ', &
                                                trim(input_filename%value), ' is blank. Terminating.'
                        call this_program%TerminateWithFatalError('ctffind','Blank image')
                    endif
                    if (make_micrograph_square) then
                        ! let's pad to make it square if we need to
                        micrograph_square_dimension = maxval(current_input_image%GetLogicalDimensions())
                        if (        current_input_image%GetLogicalDimension(1) .ne. micrograph_square_dimension &
                            .or.    current_input_image%GetLogicalDimension(2) .ne. micrograph_square_dimension) then
                            if (Isodd(micrograph_square_dimension)) micrograph_square_dimension = micrograph_square_dimension + 1 ! even may be nicer for fftw
                            call current_input_image_square%Allocate(dims=[micrograph_square_dimension, &
                                                                           micrograph_square_dimension,1])
                            call current_input_image%ClipInto(current_input_image_square, &
                                                              padding_value_real=current_input_image%GetAverageOfValues())
                            current_input_image = current_input_image_square
                            call current_input_image_square%Deallocate()
                        endif
                    endif
                    !
                    if (current_frame_within_average .eq. 1) then
                        call temp_image%Allocate(mould=current_input_image)
                        temp_image = 0.0e0
                    endif
                    call temp_image%AddImage(current_input_image)
                enddo ! end of loop over frames to average
                current_input_image = temp_image
                if (temp_image%IsAllocated()) call temp_image%Deallocate()

                ! Taper the edges of the micorgraph to lessen Gibbs artefacts
                if (taper_micrograph) then
                    call current_input_image%TaperEdges()
                    if (debug) call current_input_image%WriteToDisk('dbg_tapered.mrc')
                endif
                number_of_tiles_used = number_of_tiles_used + 1

                call current_input_image%ComputeAmplitudeSpectrumFull2D(current_power_spectrum,compute_square=fit_squared_ctf, &
                                                                        preserve_scaling=.false.) ! save time by not scaling the FFT - we will do it later
                current_power_spectrum%real_values( current_power_spectrum%physical_address_of_box_center(1),&
                                                    current_power_spectrum%physical_address_of_box_center(2),1) = 0.0e0
                if (debug) call current_power_spectrum%WriteToDisk('dbg_ampl_spectrum_before_resampling.mrc')

                ! Resample the amplitude/power spectrum
                resampling_is_necessary =    current_power_spectrum%GetLogicalDimension(1) .ne. box_size%value &
                                        .or. current_power_spectrum%GetLogicalDimension(2) .ne. box_size%value
                if (resampling_is_necessary) then
                    call current_power_spectrum%ForwardFFT(scale=.false.)
                    call resampled_power_spectrum%Allocate(dims=[box_size%value,box_size%value,1],in_real_space=.false.)
                    call current_power_spectrum%ClipInto(resampled_power_spectrum)
                    call resampled_power_spectrum%BackwardFFT(scale=.false.)
                else
                    resampled_power_spectrum = current_power_spectrum
                endif

                if (debug) call resampled_power_spectrum%WriteToDisk('dbg_ampl_spectrum_before_adding.mrc')

                call average_spectrum%AddImage(resampled_power_spectrum)

            enddo ! movie frames (tile Z) current_first_frame_within_average

            ! We need to take care of the scaling of the FFTs, as well as the averaging of tiles
            if (resampling_is_necessary) then
                call average_spectrum%MultiplyByConstant(1.0e0 /( &
                                                          real(number_of_tiles_used) &
                                                        * product(current_input_image%GetLogicalDimensions()) &
                                                        * product(current_power_spectrum%GetLogicalDimensions()) &
                                                        ))
            else
                call average_spectrum%MultiplyByConstant(1.0e0/( &
                                                          real(number_of_tiles_used) &
                                                        * product(current_input_image%GetLogicalDimensions()) &
                                                        ))
            endif
        endif ! end of test of whether we were given amplitude spectra on input


        ! Initialise the CTF object
        call current_ctf%Init(  acceleration_voltage%value,spherical_aberration%value,amplitude_contrast%value, &
                                minimum_defocus%value,minimum_defocus%value,0.0e0, &
                                1.0e0/minimum_resolution%value,1.0e0/maximum_resolution%value, &
                                astigmatism_tolerance%value,pixel_size%value, &
                                minimum_additional_phase_shift%value)
        call current_ctf%SetDefocus(minimum_defocus%value/pixel_size%value,minimum_defocus%value/pixel_size%value,0.0e0)
        call current_ctf%SetAdditionalPhaseShift(minimum_additional_phase_shift%value)


        if (.not. filtered_amplitude_spectrum_input) then
            ! Try to weaken cross artefacts
            if (debug) call average_spectrum%WriteToDisk('dbg_average_spectrum_before_cross_removal.mrc')
            call average_spectrum%ComputeAverageAndSigmaOfValuesInSpectrum(                                         &
                                    average_spectrum%GetLogicalDimension(1)*pixel_size%value/minimum_resolution%value, &
                                    average_spectrum%GetLogicalDimension(1)*1.0,           &
                                    average,sigma, &
                                    cross_half_width=12.0)
            call average_spectrum%SetMaximumValueOnCentralCross(average+10.0*sigma)


            ! Compute low-pass filtered version of spectrum
            if (subtract_background) then
                convolution_box_size = int( average_spectrum%GetLogicalDimension(1) & ! From 4.0.6
                                               * pixel_size%value &
                                               / minimum_resolution%value &
                                               * sqrt(2.0e0))
                if (IsEven(convolution_box_size)) convolution_box_size = convolution_box_size + 1
                if (debug) call average_spectrum%WriteToDisk('dbg_average_spectrum_before_conv.mrc')
                call average_spectrum%SpectrumBoxConvolution(convolution_box_size,current_power_spectrum, &
                                                             minimum_radius=average_spectrum%logical_dimensions(1) &
                                                                            *pixel_size%value/minimum_resolution%value)


                if (debug) call current_power_spectrum%WriteToDisk('dbg_background.mrc')

                ! Subtract low-pass filtered power spectrum from power spectrum. This should remove background slope.
                average_spectrum%real_values(1:average_spectrum%GetLogicalDimension(1),:,:) &
                =   average_spectrum%real_values(1:average_spectrum%GetLogicalDimension(1),:,:)&
                -   current_power_spectrum%real_values(1:average_spectrum%GetLogicalDimension(1),:,:)
            endif

            if (debug) call average_spectrum%WriteToDisk('dbg_average_spectrum_background_subtracted.mrc')


            ! Threshold high values
            threshold = average_spectrum%GetMaximumValue(minimum_distance_from_center=3,minimum_distance_from_edge=3)
            call average_spectrum%SetMaximumValue(threshold)

            ! Debug dump
            if (debug) call average_spectrum%WriteToDisk('dbg_av_spec.mrc')
        endif ! end of test for filtered amplitude spectrum input



        ! We now have a power spectrum which we can use to fit CTFs


        ! Set up the comparison object
        comparison_object%ctf = current_ctf
        comparison_object%img = average_spectrum
        comparison_object%pixel_size = pixel_size%value
        comparison_object%find_phase_shift = find_additional_phase_shift%value

        ! Do a brute-force search for the CTF parameters
        if (old_school_input .or. debug) then
            write(*,'(/a)') ' SEARCHING CTF PARAMETERS...'
        endif
        if (find_astigmatism_angle_using_mirror) then
            ! First let's look for the astigmatism angle
            temp_image = average_spectrum
            call temp_image%ApplyMirror(2)
            call average_spectrum%SearchForRotationalAlignmentWith(temp_image,45.0,5.0,estimated_astigmatism_angle, &
                                                          minimum_radius=pixel_size%value/minimum_resolution%value, &
                                                          maximum_radius=pixel_size%value/maximum_resolution%value)
            estimated_astigmatism_angle = estimated_astigmatism_angle * 0.5e0
            if (debug) print *, 'Estimated astigmatism angle = ', estimated_astigmatism_angle

            ! Now we can look for the defocus values
            bf_halfrange = [real(0.5*(maximum_defocus%value-minimum_defocus%value),kind=8), &
                            real(0.5*(maximum_defocus%value-minimum_defocus%value),kind=8), &
                            0.0d0,&
                            real(0.5*(maximum_additional_phase_shift%value-minimum_additional_phase_shift%value),kind=8) &
                            ]
            bf_midpoint  = [real(minimum_defocus%value,kind=8), &
                            real(minimum_defocus%value,kind=8), &
                            real(estimated_astigmatism_angle,kind=8), &
                            real(minimum_additional_phase_shift%value,kind=8)]
            bf_midpoint  = bf_midpoint + bf_halfrange
            bf_stepsize  = [real(defocus_search_step%value,kind=8), & ! df1
                            real(defocus_search_step%value,kind=8), & ! df2
                            0.0d0,&                                   ! angast
                            real(additional_phase_shift_search_step%value,kind=8) &
                            ]
        else

            bf_halfrange = [real(0.5*(maximum_defocus%value-minimum_defocus%value),kind=8), &
                            real(0.5*(maximum_defocus%value-minimum_defocus%value),kind=8), &
                            45.0d0,&
                            real(0.5*(maximum_additional_phase_shift%value-minimum_additional_phase_shift%value),kind=8) &
                            ]
            bf_midpoint  = [real(minimum_defocus%value,kind=8), &
                            real(minimum_defocus%value,kind=8), &
                            0.0d0, &
                            real(minimum_additional_phase_shift%value,kind=8)]
            bf_midpoint  = bf_midpoint + bf_halfrange
            bf_stepsize  = [real(defocus_search_step%value,kind=8), &
                            real(defocus_search_step%value,kind=8), &
                            5.0d0, &
                            real(additional_phase_shift_search_step%value,kind=8)]

        endif

        ! Actually run the BF search
        if (find_additional_phase_shift%value) then
            number_of_search_dimensions = 4
        else
            number_of_search_dimensions = 3
        endif
#ifdef _OPENMP
        if (.not. allocated(bf_param_array)) allocate(bf_param_array(omp_get_max_threads()))
#else
        if (.not. allocated(bf_param_array)) allocate(bf_param_array(1))
#endif
        bf_param_array = c_loc(comparison_object)
        call brute_force_search%Init(ctffind_objective_function,bf_param_array,number_of_search_dimensions,    &
                                     bf_midpoint(1:number_of_search_dimensions), &
                                     bf_halfrange(1:number_of_search_dimensions), &
                                     bf_stepsize(1:number_of_search_dimensions), &
                                     print_progress_bar=number_of_micrographs .eq. 1)
        call brute_force_search%Run()
        cg_starting_point = brute_force_search%GetBestValues()

        ! Remember the results of the search
        call current_ctf%SetDefocus(real(cg_starting_point(1))/pixel_size%value, &
                                    real(cg_starting_point(2))/pixel_size%value, &
                                    convert(real(cg_starting_point(3)),degrees,radians))
        if (find_additional_phase_shift%value) then
            call current_ctf%SetAdditionalPhaseShift(real(cg_starting_point(4)))
        endif
        call current_ctf%EnforceConvention()

        ! Print out results of brute force search
        if (old_school_input .or. debug) then
            write(*,'(a)')  '      DFMID1      DFMID2      ANGAST          CC'
            write(*,'(3(f12.2),f12.5,a)')   current_ctf%GetDefocus1InAngstroms(pixel_size%value), &
                                            current_ctf%GetDefocus2InAngstroms(pixel_size%value), &
                                            current_ctf%GetAstigmatismAzimuthInDegrees(), &
                                            -brute_force_search%GetBestScore(), &
                                             '  '
            if (debug) write(*,'(a,f0.3,a,f0.2,a)') '**debug(ctffind): found the following phase shift: ', &
                                                    current_ctf%GetAdditionalPhaseShift(), &
                                                    ' (', current_ctf%GetAdditionalPhaseShift()/3.1415,' pi)'
        endif

        ! Now we refine in the neighbourhood by using Powell's conjugate gradient algorithm
        if (old_school_input .or. debug) then
            write(*,'(/a)') ' REFINING CTF PARAMETERS...'
            write(*,'(a)')  '      DFMID1      DFMID2      ANGAST          CC'
        endif
        if (find_additional_phase_shift%value) then
            cg_accuracy = [100.0d0,100.0d0,0.5d0,0.05d0]
        else
            cg_accuracy = [100.0d0,100.0d0,0.5d0]
        endif
        call conjugate_gradient%Init(number_of_search_dimensions,ctffind_objective_function,c_loc(comparison_object), &
                                    cg_starting_point, &
                                    cg_accuracy)
        call conjugate_gradient%Run()

        ! Remember the results of the refinement
        cg_starting_point = conjugate_gradient%GetBestValues()
        call current_ctf%SetDefocus(real(cg_starting_point(1))/pixel_size%value, &
                                    real(cg_starting_point(2))/pixel_size%value, &
                                    convert(real(cg_starting_point(3)),degrees,radians))
        if (find_additional_phase_shift%value) then
            call current_ctf%SetAdditionalPhaseShift(real(cg_starting_point(4)))
        endif
        call current_ctf%EnforceConvention()


        ! Print results out to the terminal
        if (old_school_input .or. debug) then
            write(*,'(3(f12.2),f12.5,a/)')   current_ctf%GetDefocus1InAngstroms(pixel_size%value),              &
                                            current_ctf%GetDefocus2InAngstroms(pixel_size%value),               &
                                            current_ctf%GetAstigmatismAzimuthInDegrees(),                       &
                                            -conjugate_gradient%GetBestScore(),                                 &
                                            '  Final Values'
            if (debug) write(*,'(a,f0.3,a,f0.2,a)') '**debug(ctffind): found the following phase shift: ', &
                                                    current_ctf%GetAdditionalPhaseShift(), &
                                                    ' (', current_ctf%GetAdditionalPhaseShift()/3.1415,' pi)'
        endif


        ! Generate diagnostic image
        current_output_location = current_micrograph_number
        call average_spectrum%AddConstant((-1.0)*average_spectrum%GetAverageOfValuesOnEdges())
        call average_spectrum%ComputeAverageAndSigmaOfValuesInSpectrum(                                         &
                                                            current_ctf%ComputeFrequencyOfAZero(0.0,2)          &
                                                            *average_spectrum%GetLogicalDimension(1),           &
                                                            max(current_ctf%GetHighestFrequencyForFitting(),    &
                                                            current_ctf%ComputeFrequencyOfAZero(0.0,3))         &
                                                            *average_spectrum%GetLogicalDimension(1),           &
                                                            average,sigma)
        call average_spectrum%ApplyCircularMask(5.0,inverse=.true.)
        call average_spectrum%SetMaximumValueOnCentralCross(average)
        call average_spectrum%SetMinimumAndMaximumValue(average-4.0*sigma,average+4.0*sigma)
        call average_spectrum%ComputeAverageAndSigmaOfValuesInSpectrum(                                         &
                                                            current_ctf%ComputeFrequencyOfAZero(0.0,2)          &
                                                            *average_spectrum%GetLogicalDimension(1),           &
                                                            max(current_ctf%GetHighestFrequencyForFitting(),    &
                                                            current_ctf%ComputeFrequencyOfAZero(0.0,3))         &
                                                            *average_spectrum%GetLogicalDimension(1),           &
                                                            average,sigma)
        call average_spectrum%AddConstant(-1.0*average)
        call average_spectrum%MultiplyByConstant(1.0e0/sigma)
        call average_spectrum%AddConstant(average)
        call average_spectrum%Compute1DRotationalAverage(rotational_average)
        if (debug) then
            call average_spectrum%WriteToDisk('dbg_average_spectrum_before_rescaling.mrc')
        endif
        if (compute_extra_stats .or. boost_ring_contrast) then
            call average_spectrum%ComputeRotationalAverageOfPowerSpectrum( &
                                                                          current_ctf,spatial_frequency,            &
                                                                          rotational_average_astig,                 &
                                                                          rotational_average_astig_fit,             &
                                                                          frc_of_fit,frc_of_fit_sigma,              &
                                                                          rescale_input=boost_ring_contrast,        &
                                                                          squared_ctf_was_fit=fit_squared_ctf)
        endif

        call average_spectrum%ComputeAverageAndSigmaOfValuesInSpectrum(                                         &
                                                            current_ctf%ComputeFrequencyOfAZero(0.0,2)          &
                                                            *average_spectrum%GetLogicalDimension(1),           &
                                                            max(current_ctf%GetHighestFrequencyForFitting(),    &
                                                            current_ctf%ComputeFrequencyOfAZero(0.0,3))         &
                                                            *average_spectrum%GetLogicalDimension(1),           &
                                                            average,sigma)

        call average_spectrum%SetMinimumAndMaximumValue(average-1.0*sigma,average+2.0*sigma)
        if (debug) call average_spectrum%WriteToDisk('dbg_average_spectrum_before_overlay.mrc')
        call average_spectrum%OverlayCTF(current_ctf,squared_ctf=fit_squared_ctf)
        call average_spectrum%WriteToDisk(output_diagnostic_filename%value,current_output_location)

        ! Until what frequency were CTF rings detected?
        if (compute_extra_stats) then
            do last_bin_with_good_fit=2,size(frc_of_fit)
                if ( (count(frc_of_fit(1:last_bin_with_good_fit-1) .gt. 0.20d0) .gt. 3 .and. &
                           (frc_of_fit(last_bin_with_good_fit) .le. 0.2d0)) .or. &
                     (count(frc_of_fit(1:last_bin_with_good_fit-1) .gt. frc_significance_threshold) .gt. 3 .and. &
                                    ((frc_of_fit(last_bin_with_good_fit) .lt. frc_significance_threshold &
                                    .and. rotational_average_astig(last_bin_with_good_fit) .gt. 2.0d0) &
                                .or. (frc_of_fit(last_bin_with_good_fit) .lt. frc_significance_threshold &
                                    .and. rotational_average_astig(last_bin_with_good_fit) .lt. -2.0d0))&
                                                                                    )) then
                    ! last_bin_with_good_fit will now be set to point to the last frequencies at which Thon rings were still well fit
                    exit
                endif
            enddo
            last_bin_with_good_fit = min(last_bin_with_good_fit,size(frc_of_fit))
        else
            last_bin_with_good_fit = 1
        endif

        ! Print more detailled results to terminal
        if (number_of_micrographs .eq. 1) then
            write(*,'(a,f0.2,1x,a,1x,f0.2,a)')  'Estimated defocus values        : ', &
                                            current_ctf%GetDefocus1InAngstroms(pixel_size%value), ',', &
                                            current_ctf%GetDefocus2InAngstroms(pixel_size%value), ' Angstroms'
            write(*,'(a,f0.2,a)')               'Estimated azimuth of astigmatism: ', &
                                            current_ctf%GetAstigmatismAzimuthInDegrees(), ' degrees'
            if (find_additional_phase_shift%value) then
                write(*,'(a,f0.3,a,f0.2,a)')    'Additional phase shift          : ', &
                                                    current_ctf%GetAdditionalPhaseShift(), &
                                                    ' (', current_ctf%GetAdditionalPhaseShift()/3.1415,' pi)'
            endif
            write(*,'(a,f0.5)')                 'Score                           : ', &
                                            -conjugate_gradient%GetBestScore()
            if (compute_extra_stats) then
                write(*,'(a,f0.1,a)')               'Thon rings with good fit up to  : ', &
                                                pixel_size%value/spatial_frequency(last_bin_with_good_fit), ' Angstroms'
            endif
        endif


        ! Write out results to summary file
        values_to_write_out(1) = real(current_micrograph_number)
        values_to_write_out(2:4) = current_ctf%GetDefocusParametersInAngstromsAndDegrees(pixel_size%value)
        values_to_write_out(5) = current_ctf%GetAdditionalPhaseShift()
        values_to_write_out(6) = -conjugate_gradient%GetBestScore()
        if (compute_extra_stats) then
            values_to_write_out(7) = pixel_size%value/spatial_frequency(last_bin_with_good_fit)
        else
            values_to_write_out(7) = 0.0
        endif
        call output_text%WriteDataLine(values_to_write_out)



        ! Write avrot
        if (compute_extra_stats) then
            if (output_text_avrot%number_of_data_lines .eq. 0) then
                call output_text_avrot%Init(output_text_fn,OPEN_TO_WRITE,size(rotational_average))
                call output_text_avrot%WriteCommentLine(comment_lines(1))
                call output_text_avrot%WriteCommentLine(comment_lines(2))
                call output_text_avrot%WriteCommentLine(comment_lines(3))
                call output_text_avrot%WriteCommentLine(comment_lines(4))
                call output_text_avrot%WriteCommentLine('6 lines per micrograph: #1 - spatial frequency (1/pixels); '//&
                                                 '#2 - 1D rotational average of spectrum (assuming no astigmatism); '//&
                                                 '#3 - 1D rotational average of spectrum; #4 - CTF fit; '//&
                                                 '#5 - cross-correlation between spectrum and CTF fit; '//&
                                                 '#6 - 2sigma of expected cross correlation of noise')
            endif
            call output_text_avrot%WriteDataLine(real(spatial_frequency))
            call output_text_avrot%WriteDataLine(real(rotational_average))
            call output_text_avrot%WriteDataLine(real(rotational_average_astig))
            call output_text_avrot%WriteDataLine(real(rotational_average_astig_fit))
            call output_text_avrot%WriteDataLine(real(frc_of_fit))
            call output_text_avrot%WriteDataLine(real(frc_of_fit_sigma))
        endif

        ! Mark progress
        if (.not. old_school_input .and. number_of_micrographs .gt. 1) then
            call my_progress_bar%Update(current_micrograph_number)
        endif

    enddo ! end of loop over micrographs

    if (.not. old_school_input .and. number_of_micrographs .gt. 1) then
        call my_progress_bar%Finish()
    endif

    ! Tell the user where the outputs are
    write(*,'(/2a)')'Summary of results                          : ', trim(adjustl(output_text%filename))
    write(*,'(2a)') 'Diagnostic images                           : ', trim(adjustl(output_diagnostic_filename%value))
    if (compute_extra_stats) then
        write(*,'(2a)') 'Detailled results, including 1D fit profiles: ', trim(adjustl(output_text_avrot%filename))
        write(*,'(2a)') 'Use this command to plot 1D fit profiles    : ctffind_plot_results.sh ', &
                                                                      trim(adjustl(output_text_avrot%filename))
    endif



    ! Done running the program
    call this_program%Terminate()

end program ctffind
